@@ -6,6 +6,7 @@ import websockets
 import json
 import urllib
 import threading
+from concurrent.futures._base import CancelledError
 from dataclasses import asdict
 from aiohttp import web
 from rpc_gateway import messages, errors
@@ -36,7 +37,7 @@ class Gateway:
     SERVER_MESSAGES = (messages.Method.GET, messages.Method.SET, messages.Method.CALL, messages.Method.LOCK,
                        messages.Method.UNLOCK, messages.Method.METADATA)
 
-    def __init__(self, host: str = 'localhost', port: int = 8888, http_port: int = 8887, auth_key: str = 'DEFAULT_KEY'):
+    def __init__(self, host: str = 'localhost', port: int = 8888, http_port: int = 8887, auth_key: str = 'DEFAULT_KEY', event_loop: Optional[asyncio.AbstractEventLoop] = None):
         self.host = host
         self.port = port
         self.http_port = http_port
@@ -44,17 +45,22 @@ class Gateway:
         self.logger = logger.getChild(self.__class__.__name__)
         self.websocket_connections: List[WebsocketConnection] = []
         self.websocket: Optional[websockets.WebSocketServer] = None
-        self.event_loop = asyncio.get_event_loop()
+        self.event_loop = event_loop or asyncio.get_event_loop()
         self.instances: Dict[str, WebsocketConnection] = {}
         self.instance_groups: Dict[str, List[str]] = {}
         self.websocket_instances: Dict[WebsocketConnection, List[str]] = {}
+        self.check_connections_task: Optional[asyncio.Task] = None
+        self.running = False
 
         self.http_server: Optional[web.Server] = None
         self.http_runner: Optional[web.ServerRunner] = None
         self.http_site: Optional[web.TCPSite] = None
 
     def start(self, wait = True):
-        await_sync(self._start(wait), self.event_loop)
+        self.event_loop.run_until_complete(self._start(wait=False))
+
+        if wait:
+            await_sync(self._wait())
 
     async def _start(self, wait = True):
         self.logger.info(f'Starting on ws://{self.host}:{self.port}')
@@ -66,6 +72,8 @@ class Gateway:
         await self.http_runner.setup()
         self.http_site = web.TCPSite(self.http_runner, self.host, self.http_port)
         await self.http_site.start()
+        self.running = True
+        self.check_connections_task = self.check_connections()
 
         if wait:
             await self._wait()
@@ -83,12 +91,14 @@ class Gateway:
         await_sync(self._stop(), self.event_loop)
 
     async def _stop(self):
+        self.running = False
         await self.http_site.stop()
         try:
             await asyncio.gather(*[websocket_connection.stop() for websocket_connection in self.websocket_connections])
         except asyncio.exceptions.CancelledError:
             pass
         self.websocket.close()
+        await self.check_connections_task
 
     async def on_connection(self, connection: websockets.WebSocketServerProtocol, path: str):
         self.logger.info(f'New connection from {connection.remote_address} path: {path}')
@@ -104,13 +114,6 @@ class Gateway:
 
         await websocket_connection.start()
 
-        if websocket_connection in self.websocket_instances:
-            for instance in self.websocket_instances[websocket_connection]:
-                self.instances.pop(instance)
-        self.websocket_connections.remove(websocket_connection)
-
-        self.logger.info(f'Connection from {connection.remote_address} closed')
-
     async def _unlock_instance(self, instance_name: str):
         server = self.instances[instance_name]
         return await server.request(messages.Request(method=messages.Method.UNLOCK, data=instance_name), raise_error=False)
@@ -121,6 +124,14 @@ class Gateway:
 
         return instance in self.instance_groups[group]
 
+    async def check_connections(self):
+        while self.running:
+            for websocket_connection in self.websocket_connections:
+                if not websocket_connection.connected:
+                    await self.handle_close(websocket_connection)
+
+            time.sleep(1)
+
     #
     # Request Handlers
     #
@@ -164,6 +175,12 @@ class Gateway:
 
         # check to see if this instance
         for instance_name, instance_group in request.data:
+            if instance_name is None:
+                return messages.Response(status=messages.Status.ERROR, data=errors.InvalidMessageError(f'Cannot register instance with name: {instance_name}'))
+
+            if instance_group is None:
+                return messages.Response(status=messages.Status.ERROR, data=errors.InvalidMessageError(f'Cannot register instance with group: {instance_group}'))
+
             if self.is_registered(instance_group, instance_name):
                 return messages.Response(status=messages.Status.ERROR, data=errors.InstanceAlreadyRegisteredError(f'Instance "{instance_name}" already registered with "{instance_group}" group'))
 
@@ -177,6 +194,24 @@ class Gateway:
 
         return messages.Response()
 
+    async def handle_deregister_request(self, websocket_connection: WebsocketConnection, request: messages.Request) -> messages.Response:
+        self.logger.info(f'Deregistering instances: {request.data}')
+
+        # find instance to deregister
+        for instance_name in request.data:
+            if instance_name not in self.instances:
+                return messages.Response(status=messages.Status.ERROR, data=errors.InstanceNotFoundError(f'Instance "{instance_name}" not found'))
+
+            instance_connection = self.instances[instance_name]
+            self.instances.pop(instance_name)
+            self.websocket_instances[instance_connection].remove(instance_name)
+
+            for group, group_instances in self.instance_groups.items():
+                if instance_name in group_instances:
+                    group_instances.remove(instance_name)
+
+        return messages.Response()
+
     # this is called by the GatewayConnection MessagePump when a new request is received
     async def handle_request(self, websocket_connection: WebsocketConnection, request: messages.Request) -> messages.Response:
         if request.method in self.SERVER_MESSAGES:
@@ -188,6 +223,9 @@ class Gateway:
         if request.method == messages.Method.REGISTER:
             return await self.handle_register_request(websocket_connection, request)
 
+        if request.method == messages.Method.DEREGISTER:
+            return await self.handle_deregister_request(websocket_connection, request)
+
         if request.method == messages.Method.LIST:
             return await self.handle_list_request(request)
 
@@ -218,6 +256,10 @@ class Gateway:
         for request_id, response_queue in websocket_connection.receive_queues.items():
             await response_queue.put(messages.Response(status=messages.Status.ERROR, data=errors.ServerConnectionLostError(f'Server connection lost')))
 
+        self.websocket_connections.remove(websocket_connection)
+
+        self.logger.info(f'Connection from {websocket_connection.connection.remote_address} closed')
+
     def check_authentication(self, request: web.BaseRequest) -> bool:
         if 'Authentication' in request.headers:
             auth_header = request.headers['Authentication']
@@ -315,11 +357,19 @@ class GatewayClient:
     async def _on_start(self):
         pass
 
+    async def _on_stop(self):
+        pass
+
     def stop(self):
-        await_sync(self._stop(), self.event_loop)
+        try:
+            await_sync(self._stop(), self.event_loop)
+        except CancelledError:
+            pass
+
         self.start_thread.join()
 
     async def _stop(self):
+        await self._on_stop()
         await self.websocket_connection.stop()
 
     async def _log_and_handle_request(self, websocket_connection: WebsocketConnection, request: messages.Request) -> messages.Response:
