@@ -8,4 +8,7 @@ def await_sync(awaitable: Awaitable, loop: Optional[asyncio.AbstractEventLoop] =
         future = asyncio.run_coroutine_threadsafe(awaitable, loop or asyncio.get_event_loop())
         return future.result(timeout=timeout)
     else:
-        return loop.run_until_complete(awaitable)
+        try:
+            return loop.run_until_complete(awaitable)
+        except asyncio.CancelledError:
+            pass
